import os
import re
import shutil
import subprocess
import yaml

from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv
from charmhelpers.core import host
from charmhelpers.core.templating import render
from charms.reactive import hook, set_flag, clear_flag, when, when_all, when_not
from charms import apt
from charms.layer import status

import influxdb

from influxdbcharmlib import create_influxdb_user, get_influxdb_credentials  # NOQA: E402


CRONFILE = '/etc/cron.daily/influxdb-charm-backup'
CHECK_SRC = 'files/check_influxdb.py'
CHECK_DIR = '/usr/local/lib/nagios/plugins'
CHECK_CMD = CHECK_DIR + '/check_influxdb.py'
INFLUXDB_SERVICE = "influxdb.service"


@when('admin.available')
def configure_website(website):
    config = hookenv.config()
    port = config.get('bind_port')
    website.configure(port=port)


@when('apt.installed.influxdb')
@when_not('influxdb.has_client')
def install_influx_client():
    # This is a hack because the CLI shell tool is packaged differently in
    # upstream and Debian/Ubuntu. Upstream has no 'influxdb-client' package.
    if not shutil.which('influx'):
        apt.queue_install('influxdb-client')
        apt.install_queued()
    set_flag('influxdb.has_client')


@when_not('influxdb.configured')
@when('influxdb.has_client')
def install_influx():
    config = hookenv.config()
    port = config.get('bind_port')
    hookenv.open_port(8083)
    hookenv.open_port(port)
    config_changed()
    host.service_start('influxdb')
    hookenv.status_set('active', '')
    set_flag('influxdb.configured')
    set_flag('config.changed')
    configure_influxdb_auth()


def configure_influxdb_auth():
    # Create admin account, if it doesn't already exist.
    username, password = get_influxdb_credentials()
    if not username:
        username = 'admin'
        create_influxdb_user(username, admin=True, local_user='root')
        set_flag('influxdb.{}.account.created'.format(username))


def apply_configs(fh, config):  # noqa: TODO: fix overly complex
    new_config = []

    section = 'global'
    indent = ''

    section_keys_found = {
        'global': {},
    }
    initial_comments = True
    for line in fh:
        # Start after initial comments "Welcome to the InfluxDB configur..."
        if initial_comments and not line.startswith('##'):
            initial_comments = False
        if line.startswith('[') or (not initial_comments and line.startswith('##')):
            if section in config:
                # If the configs weren't found in this section let's add them.
                for key in config[section].keys():
                    if key in section_keys_found[section]:
                        break
                    section_keys_found[section][key] = ''
                    new_config.append("{}{} = {}\n\n".format(indent, key, config[section][key]))
            new_config.append(line)
            if line.startswith('['):
                section = line[1:][:-2]
                # Apart from global config options, indent the rest.
                indent = '  '
            continue
        # Format is 'key = value' or '# key = value'
        match = re.match(r'^([\s#]*)(\S+)(\s*=.*)', line)
        if not match:
            new_config.append(line)
            continue
        key = match.group(2)
        value = match.group(3)
        # Store all config options we found per section.
        if section not in section_keys_found:
            section_keys_found[section] = {}
        section_keys_found[section][key] = ''
        if section not in config or key not in config[section]:
            # We want to keep {meta,data}.{dir,wal-dir} always.
            if section in ['meta', 'data'] and key in ['dir', 'wal-dir']:
                new_config.append(line)
            else:
                new_config.append("{}# {}{}\n".format(indent, key, value))
            continue
        new_config.append("{}{} = {}\n".format(indent, key, config[section][key]))

    return ''.join(new_config)


@when_all('influxdb.configured', 'influxdb.admin.account.created', 'config.changed')
def config_changed():
    config = hookenv.config()
    conf = {
        'global': {},
    }

    # Only write out the bind-address if it differs from the default.
    if config.get('ip_address') or config.get('bind_port') != 8086:
        conf['http'] = {}
        conf['http']['bind-address'] = '"{}:{}"'.format(config.get('ip_address', ''),
                                                        config.get('bind_port'))

    # Construct our dict of configs from the extra_configs charm option.
    if config.get('extra_configs', None):
        for c in yaml.safe_load(config.get('extra_configs', '[]')):
            k, value = c.split('=', 1)
            section, key = k.split('.', 1)
            if section not in conf:
                conf[section] = {}
            if str(value) in ('true', 'false') or value.isdigit():
                conf[section][key] = value
            else:
                conf[section][key] = '"{}"'.format(value)

    target = '/etc/influxdb/influxdb.conf'
    if not os.path.exists(target):
        shutil.copy2('files/influxdb.conf', target)
    with open(target, 'r') as f:
        updated_config = apply_configs(f, conf)
    host.write_file(target, updated_config.encode("utf-8"), perms=0o644)

    # Write the systemd override timeout for the start of the service
    if host.service_available(INFLUXDB_SERVICE):
        systemd_override_dir = "/etc/systemd/system/{0}.d/".format(INFLUXDB_SERVICE)
        host.mkdir(systemd_override_dir)
        systemd_override = os.path.join(systemd_override_dir, "override.conf")
        shutil.copy2("files/influxdb_override.conf", systemd_override)
        hookenv.log("Setting TimeoutStartSec override for influxdb service", hookenv.DEBUG)
        try:
            subprocess.check_call(["systemctl", "daemon-reload"])
        except (OSError, subprocess.CalledProcessError):
            hookenv.log("Cannot reload systemd", hookenv.ERROR)
            status.blocked("Cannot configure influxdb.service")
    hookenv.log("Restarting influxdb service", hookenv.INFO)
    host.service_stop('influxdb')
    host.service_start('influxdb')


@when('config.changed.backup_dir')
def create_backup_job():
    config = hookenv.config()
    backup_dir = config.get('backup_dir')
    if backup_dir:
        render(source='influxdb-backup',
               target=CRONFILE,
               perms=0o755,
               context={
                   'backup_dir': backup_dir,
                   'backup_retention': 3,
               })
    else:
        old_backup_dir = config.previous('backup_dir')
        hookenv.log('NOT removing previous backup directory {}'.format(old_backup_dir))
        try:
            os.remove(CRONFILE)
            hookenv.log('Removed {}'.format(CRONFILE))
        except OSError as e:
            hookenv.log('Cannot remove {} - ignoring'.format(CRONFILE), hookenv.ERROR)


@hook('upgrade-charm')
def upgrade_charm():
    # If auth is not configured correctly, block the upgrade.
    if os.path.exists('/root/.influx-auth'):
        # This could be more rigorously checked.
        set_flag('influxdb.admin.account.created')
    else:
        ic = influxdb.InfluxDBClient()
        try:
            ic.get_list_database()
        except influxdb.exceptions.InfluxDBClientError as e:
            if e.code == 403:
                # InfluxDB throws 403 if there is no admin account, but
                # will still allow it to be created, without authenticating.
                hookenv.log(
                    'Received 403 from InfluxDB, which indicates authentication is enabled '
                    'but no admin account exists.  Continuing to let it be created.')
            elif e.code == 401:
                auth_error_message = \
                    'Authentication is enabled but /root/.influx-auth does not exist. ' \
                    'Please place username and password for an admin account ' \
                    'named "admin" in this file and retry the upgrade.'
                hookenv.status_set('blocked', auth_error_message)
                hookenv.log(auth_error_message, hookenv.ERROR)
                raise
            else:
                raise

    clear_flag('influxdb.configured')
    set_flag('config.changed.backup_dir')
    # If we have an NRPE relation, force run of that as well
    if hookenv.relation_ids('nrpe-exteral-master'):
        set_flag('nrpe-external-master.available')
    hookenv.status_set('active', '')


@when('query.api.available', 'influxdb.admin.account.created')
def query_available(query):
    config = hookenv.config()
    port = config.get('bind_port')
    username, password = get_influxdb_credentials()
    query.configure(port=port, username=username, password=password)


@when('grafana-source.available')
def configure_grafana(grafana):
    config = hookenv.config()
    port = config.get('bind_port')
    grafana.provide('influxdb', port, 'Juju generated source')


@when('nrpe-external-master.available')
def update_nrpe_config(svc):
    # copy the check into place
    os.makedirs(CHECK_DIR, exist_ok=True)
    shutil.copy(CHECK_SRC, CHECK_CMD)
    os.chmod(CHECK_CMD, 0o755)
    # install the nrpe configuration
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    nrpe_setup.add_check(
        shortname='influxdb',
        description='Check InfluxDB',
        check_cmd=CHECK_CMD,
    )
    nrpe_setup.write()


@when('nrpe-external-master.available', 'influxdb.admin.account.created')
@when_not('influxdb.nagios.account.created')
def create_nagios_influxdb_user(svc):
    create_influxdb_user('nagios', admin=True, local_user='nagios')
    set_flag('influxdb.nagios.account.created')
